#ifndef DUNE_SUBGRIDGEOMETRY_HH
#define DUNE_SUBGRIDGEOMETRY_HH

/** \file
* \brief The SubGridGeometry class and its specializations
*/
#include <type_traits>

#include <dune/common/fmatrix.hh>
#include <dune/common/typetraits.hh>

namespace Dune {

template<int mydim, int coorddim, class GridImp> class SubGridGeometry;
template<int mydim, int coorddim, class GridImp> class SubGridLocalGeometry;

/** \brief Wrapping class for host grid Geometries and LocalGeometries

    \tparam isLocal If true we are wrapping a LocalGeometry

    I cannot put the isLocal parameter directly in the SubGridGeometry class,
    because the Dune grid interface specifies the number and types of template
    parameters of that class.
*/
template<int mydim, int coorddim, class GridImp, bool isLocal>
class SubGridGeometryBase
{
    private:

        typedef typename GridImp::ctype ctype;


    public:

        // The codimension of this entitypointer wrt the host grid
        enum {CodimInHostGrid = GridImp::HostGridType::dimension - mydim};
        enum {DimensionWorld = GridImp::HostGridType::dimensionworld};

        // Select the type we are wrapping depending on the isLocal flag
        // - isLocal == true:  LocalGeometry
        // - isLocal == false: Geometry
        typedef typename std::conditional<isLocal,
                                    typename GridImp::HostGridType::Traits::template Codim<CodimInHostGrid>::LocalGeometry,
                                    typename GridImp::HostGridType::Traits::template Codim<CodimInHostGrid>::Geometry>::type HostGridGeometry;

        //! The type of the Jacobian matrix of the mapping from the reference element to this element
        typedef typename HostGridGeometry::JacobianTransposed JacobianTransposed;
        //! The type of the nverse of the Jacobian matrix of the mapping from the reference element to this element
        typedef typename HostGridGeometry::JacobianInverseTransposed JacobianInverseTransposed;

        /** Constructor for a given host grid geometry
        */
        SubGridGeometryBase(const HostGridGeometry& hostGeometry)
            : hostGeometry_(hostGeometry)
        {
        }


        /** \brief Return the element type identifier
        */
        GeometryType type () const {
            return hostGeometry_.type();
        }


        /** \brief Return true if the geometry mapping is affine and false otherwise */
        bool affine() const {
            return hostGeometry_.affine();
        }

        //! return the number of corners of this element. Corners are numbered 0...n-1
        int corners () const {
            return hostGeometry_.corners();
        }


        //! access to coordinates of corners. Index is the number of the corner
        FieldVector<ctype, coorddim> corner (int i) const {
            return hostGeometry_.corner(i);
        }


        /** \brief Maps a local coordinate within reference element to
        * global coordinate in element  */
        FieldVector<ctype, coorddim> global (const FieldVector<ctype, mydim>& local) const{
            return hostGeometry_.global(local);
        }


        /** \brief Maps a global coordinate within the element to a
        * local coordinate in its reference element */
        FieldVector<ctype, mydim> local (const FieldVector<ctype, coorddim>& global) const {
            return hostGeometry_.local(global);
        }


        /**
        */
        ctype integrationElement (const FieldVector<ctype, mydim>& local) const {
            return hostGeometry_.integrationElement(local);
        }

        /** \brief return volume of geometry */
        ctype volume () const
        {
            return hostGeometry_.volume();
        }

        /** \brief return center of geometry
        *  Note that this method is still subject to a change of name and semantics.
        *  At the moment, the center is not required to be the centroid of the
        *  geometry, or even the centroid of its corners. This makes the current
        *  default implementation acceptable, which maps the centroid of the
        *  reference element to the geometry.
        *  We may change the name (and semantic) of the method to centroid() if we
        *  find reasonably efficient ways to implement it properly.
        */
        FieldVector<ctype, coorddim> center () const
        {
            return hostGeometry_.center();
        }


        //! The Jacobian matrix of the mapping from the reference element to this element
        const JacobianTransposed jacobianTransposed (const FieldVector<ctype, mydim>& local) const {
            return hostGeometry_.jacobianTransposed(local);
        }


        //! The inverse of the Jacobian matrix of the mapping from the reference element to this element
        const JacobianInverseTransposed jacobianInverseTransposed (const FieldVector<ctype, mydim>& local) const {
            return hostGeometry_.jacobianInverseTransposed(local);
        }


        const HostGridGeometry hostGeometry_;

};

/** \brief Encapsulate a host grid Geometry */
template<int mydim, int coorddim, class GridImp>
class SubGridGeometry :
    public SubGridGeometryBase<mydim,coorddim,GridImp,false>
{
    public:
    enum {CodimInHostGrid = GridImp::HostGridType::dimension - mydim};

        // select appropiate hostgrid geometry via typeswitch
        typedef typename GridImp::HostGridType::Traits::template Codim<CodimInHostGrid>::Geometry HostGridGeometry;

        /** Constructor for a given host grid geometry
        */
        SubGridGeometry(const HostGridGeometry& hostGeometry)
            : SubGridGeometryBase<mydim,coorddim,GridImp,false>(hostGeometry)
        {}

};

/** \brief Encapsulate a host grid LocalGeometry */
template<int mydim, int coorddim, class GridImp>
class SubGridLocalGeometry :
    public SubGridGeometryBase<mydim,coorddim,GridImp,true>
{
    public:
    enum {CodimInHostGrid = GridImp::HostGridType::dimension - mydim};

        // select appropiate hostgrid geometry via typeswitch
        typedef typename GridImp::HostGridType::Traits::template Codim<CodimInHostGrid>::LocalGeometry HostGridLocalGeometry;

        /** Constructor for a given host grid local geometry
        */
        SubGridLocalGeometry(const HostGridLocalGeometry& hostGeometry)
            : SubGridGeometryBase<mydim,coorddim,GridImp,true>(hostGeometry)
        {}

};

}  // namespace Dune

#endif
