#include <config.h>

#define DISABLE_DEPRECATED_METHOD_CHECK 1

#include <dune/common/exceptions.hh>

#include <dune/grid/yaspgrid.hh>
//#include <dune/grid/sgrid.hh>
#include <dune/grid/onedgrid.hh>

#if HAVE_ALBERTA
#include <dune/grid/albertagrid.hh>
#endif

#if HAVE_ALUGRID
#include <dune/grid/albertagrid.hh>
#endif

#if HAVE_UG
#include <dune/grid/albertagrid.hh>
#endif

#include <dune/grid/utility/structuredgridfactory.hh>
#include <dune/grid/test/gridcheck.hh>
#include <dune/grid/test/checkintersectionit.hh>
#include <dune/grid/io/file/vtk/vtkwriter.hh>

#include <dune/subgrid/subgrid.hh>

using namespace Dune;

template<class GridType, bool MapIndexStorage>
struct EnableLevelIntersectionIteratorCheck< SubGrid<GridType::dimension,GridType, MapIndexStorage> >
{
  static const bool v = false;
};

enum EltGeometryType {SIMPLEX, CUBE, NA};

template <class GridType, EltGeometryType EGT>
struct CoarseGridConstructor
{
    static Dune::shared_ptr<GridType> constructCoarseGrid(size_t ncells=1)
    {
        DUNE_THROW(Dune::NotImplemented,"not implemented for this Element geometry");
    }
};

template <class GridType>
struct CoarseGridConstructor<GridType, SIMPLEX>
{
    static Dune::shared_ptr<GridType> constructCoarseGrid(size_t ncells=1)
    {
        const int dim=GridType::dimension;
        Dune::array<unsigned int, dim> elts;
        elts.fill(ncells);
        return Dune::StructuredGridFactory<GridType>::createSimplexGrid(Dune::FieldVector<double,dim>(-1.0),Dune::FieldVector<double,dim>(1.0),elts);
    }
};

template <class GridType>
struct CoarseGridConstructor<GridType, CUBE>
{
    static Dune::shared_ptr<GridType> constructCoarseGrid(size_t ncells=1)
    {
        const int dim=GridType::dimension;
        Dune::array<unsigned int, dim> elts;
        elts.fill(ncells);
        return Dune::StructuredGridFactory<GridType>::createCubeGrid(Dune::FieldVector<double,dim>(-1.0),Dune::FieldVector<double,dim>(1.0),elts);
    }
};

template<class GridType>
void refineLocally(GridType& grid, int refine=3)
{
    typedef typename GridType::LevelGridView GridView;
    typedef typename GridType::LevelGridView::template Codim<0>::Iterator ElementIterator;

    for(int i=0; i<refine; ++i)
    {
        GridView gv = grid.levelGridView(grid.maxLevel());
        int half = gv.size(0)/2 + 1;

        ElementIterator it = gv.template begin<0>();
        ElementIterator end = gv.template end<0>();
        for(; it!=end; ++it)
        {
            if (gv.indexSet().index(*it) < half)
                grid.mark(1, *it);
        }
        grid.preAdapt();
        grid.adapt();
        grid.postAdapt();
    }
}

template<class GridType>
void createPartialDomainSubgrid(GridType& subgrid, const double r1=0.5, const double r2=0.9)
{
    std::set<size_t> elementsForSubgrid;

    typedef typename GridType::HostGridType HostGridType;
    typedef typename HostGridType::LeafGridView::template Codim<0>::Iterator HostElementIterator;

    HostGridType& hostgrid = subgrid.getHostGrid();

    const typename HostGridType::Traits::GlobalIdSet& globalIDset = hostgrid.globalIdSet();

    HostElementIterator element = hostgrid.leafGridView().template begin<0>();
    HostElementIterator elementEnd = hostgrid.leafGridView().template end<0>();
    for (;element!=elementEnd; ++element)
    {
        typename HostGridType::LeafGridView::template Codim<0>::Geometry geom = element->geometry();
        for (size_t v=0; v<geom.corners(); ++v)
        {
            if (fabs(geom.corner(v)[0])<=0.2 or fabs(geom.corner(v)[1])<=0.3)
            {
                elementsForSubgrid.insert(globalIDset.template id<0>(*element));
                break;
            }
        }
    }

    subgrid.createBegin();
    subgrid.insertSetPartial(elementsForSubgrid);
    subgrid.createEnd();

    return;
}

template<class GridType>
void createRingDomainSubgrid(GridType& subgrid, const double r1=0.5, const double r2=0.9)
{
    std::set<size_t> elementsForSubgrid;

    typedef typename GridType::HostGridType HostGridType;
    typedef typename HostGridType::LeafGridView::template Codim<0>::Iterator HostElementIterator;

    HostGridType& hostgrid = subgrid.getHostGrid();

    const typename HostGridType::Traits::GlobalIdSet& globalIDset = hostgrid.globalIdSet();

    HostElementIterator element = hostgrid.leafGridView().template begin<0>();
    HostElementIterator elementEnd = hostgrid.leafGridView().template end<0>();
    for (;element!=elementEnd; ++element)
    {
        double  sum = 0.0,
                diffsum = 0.0;

        typename HostGridType::LeafGridView::template Codim<0>::Geometry geom = element->geometry();
        for (size_t v=0; v<geom.corners(); ++v)
        {
            sum += geom.corner(v).two_norm();
            for (size_t w=v+1; w<geom.corners(); ++w)
            {
                typename HostGridType::LeafGridView::template Codim<0>::Geometry::GlobalCoordinate dummy=geom.corner(w);
                dummy -= geom.corner(v);

                if (dummy.two_norm()>2.0*r2)
                {
                    elementsForSubgrid.insert(globalIDset.template id<0>(*element));
                    break;
                }
            }
        }
        if (sum > geom.corners()*r1 and sum < geom.corners()*r2)
            elementsForSubgrid.insert(globalIDset.template id<0>(*element));
    }

    subgrid.createBegin();
    subgrid.insertSetPartial(elementsForSubgrid);
    subgrid.createEnd();

    return;
}

// test subgrid for given dimension
template <class HostGridType, EltGeometryType EGT>
size_t checkWithHostGrid(size_t host_prerefine=2, int sub_initlevel=-1, size_t sub_localrefine=2, size_t host_level0_ncells=1)
{
    size_t ret = 0;
    const int dim=HostGridType::dimension;

    Dune::shared_ptr<HostGridType> hostgridptr = (CoarseGridConstructor<HostGridType,EGT>::constructCoarseGrid(host_level0_ncells));
    HostGridType& hostgrid = *hostgridptr;

//    hostgrid.setClosureType(HostGridType::NONE);

    std::cout << "Macro-Hostgrid created with " << hostgrid.size(dim) << " nodes and " << hostgrid.size(0) << " elements on level 0." << std::endl;

    for (size_t i=0; i<host_prerefine; ++i)
        hostgrid.globalRefine(1);

    std::cout << "Hostgrid uniformly refined to level " << hostgrid.maxLevel() << " with " << hostgrid.size(dim) << " nodes and " << hostgrid.size(0) << " elements." << std::endl;

    typedef SubGrid<dim,HostGridType> SubGridType;
    SubGridType subgrid(hostgrid);

    typedef typename HostGridType::template Codim<0>::LevelIterator HostElementIterator;
    typedef typename SubGridType::template Codim<0>::LevelIterator  SubElementIterator;

    // create subgrid with insertLevel
    {
        subgrid.createBegin();
        subgrid.insertLevel(hostgrid.maxLevel()+sub_initlevel);
        subgrid.createEnd();
        subgrid.setMaxLevelDifference(1);
    }

    std::cout << "Uniformly refined SubGrid created with " << subgrid.size(dim) << " nodes and " << subgrid.size(0) << " elements with maxlevel= " << subgrid.maxLevel() << "." << std::endl;

    subgrid.report();

    try {
        std::cout << "Performing gridcheck\n";
        gridcheck(subgrid);
    }
    catch (Dune::NotImplemented& e)
    {
        std::cerr << e << std::endl;
        ret = 1;
    }
    try {
        std::cout << "Performing checkIntersectionIterator\n";
        checkIntersectionIterator(subgrid);
    }
    catch (Dune::NotImplemented& e)
    {
        std::cerr << e << std::endl;
        ret = 1;
    }

#if 0
    //create subgrid with insertSet

    // check subgrid
    std::cout << "Check subgrid with dim=" << dim << std::endl;

    gridcheck(subgrid);
    checkIntersectionIterator(subgrid);

#endif
    // refine subgrid
    refineLocally(subgrid,3);
    subgrid.report();

    std::cout << "Locally refined SubGrid created with " << subgrid.size(dim) << " nodes and " << subgrid.size(0) << " elements with maxlevel= " << subgrid.maxLevel() << "." << std::endl;

    // check locally refined subgrid
    try {
        std::cout << "Performing gridcheck\n";
        gridcheck(subgrid);
    }
    catch (Dune::NotImplemented& e)
    {
        std::cerr << e << std::endl;
        ret = 1;
    }
    try {
        std::cout << "Performing checkIntersectionIterator\n";
        checkIntersectionIterator(subgrid);
    }
    catch (Dune::NotImplemented& e)
    {
        std::cerr << e << std::endl;
        ret = 1;
    }

    createPartialDomainSubgrid(subgrid);
    Dune::VTKWriter<typename SubGridType::LeafGridView> writer(subgrid.leafGridView());
    writer.write("grid");

    std::cout << "Locally refined SubGrid with domain coverage <100\% created with " << subgrid.size(dim) << " nodes and " << subgrid.size(0) << " elements with maxlevel= " << subgrid.maxLevel() << "." << std::endl;
    try {
        std::cout << "Performing gridcheck\n";
        gridcheck(subgrid);
    }
    catch (Dune::NotImplemented& e)
    {
        std::cerr << e << std::endl;
        ret = 1;
    }
    try {
        std::cout << "Performing checkIntersectionIterator\n";
        checkIntersectionIterator(subgrid);
    }
    catch (Dune::NotImplemented& e)
    {
        std::cerr << e << std::endl;
        ret = 1;
    }

    return ret;
}

